import uuid

from datetime import date, datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from sqlalchemy import (
    Boolean,
    Date,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    Numeric,
    Uuid,
    select,
)
from sqlalchemy.orm import Mapped, Session, mapped_column, relationship

from ..db.base_class import reg


if TYPE_CHECKING:
    # if the target of the relationship is in another module
    # that cannot normally be imported at runtime
    from .employee import Employee
    from .user import User


@reg.mapped_as_dataclass(unsafe_hash=True)
class Attendance:
    __tablename__ = "attendances"
    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    employee_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("employees.id"), nullable=False)
    date_: Mapped[date] = mapped_column("date", Date, nullable=False)
    attendance_type: Mapped[int] = mapped_column(Integer, nullable=False)
    amount: Mapped[Decimal] = mapped_column(Numeric(precision=5, scale=2), nullable=False)
    creation_date: Mapped[datetime] = mapped_column(DateTime(), nullable=False)
    user_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("users.id"), nullable=False)
    is_valid: Mapped[bool] = mapped_column(Boolean, nullable=False)

    __table_args__ = (
        Index(
            "only_one_valid_attendance",
            "employee_id",
            "date",
            unique=True,
            postgresql_where=(is_valid == True),  # noqa: E712
        ),
    )
    user: Mapped["User"] = relationship("User")
    employee: Mapped["Employee"] = relationship("Employee", back_populates="attendances")

    def __init__(
        self,
        employee_id: uuid.UUID,
        date_: date,
        attendance_type: int,
        user_id: uuid.UUID,
        amount: Decimal = Decimal(0),
        creation_date: datetime | None = None,
        is_valid: bool = True,
        id_: uuid.UUID | None = None,
    ):
        self.employee_id = employee_id
        self.date_ = date_
        self.attendance_type = attendance_type
        self.user_id = user_id
        self.amount = amount
        self.creation_date = creation_date or datetime.utcnow()
        self.is_valid = is_valid
        if id_ is not None:
            self.id = id_

    def create(self, db: Session) -> None:
        old = (
            db.execute(
                select(Attendance)
                .filter(Attendance.date_ == self.date_)
                .filter(Attendance.employee_id == self.employee_id)
                .filter(Attendance.is_valid == True)  # noqa: E712
            )
            .scalars()
            .one_or_none()
        )
        if old is None or old.attendance_type != self.attendance_type:
            if old is not None:
                old.is_valid = False
            db.add(self)
