from fastapi import APIRouter, Security
from sqlalchemy import select
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.db_setting import DbSetting
from ..models.setting_type import SettingType
from ..models.user import User
from ..schemas.settings import Maintenance
from ..schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=Maintenance)
def get_maintenance(
    user: UserToken = Security(get_user),
) -> Maintenance:
    with SessionFuture() as db:
        data = db.execute(select(DbSetting).where(DbSetting.name == "Maintenance")).scalars().one_or_none()
        return info(data, db)


@router.post("", response_model=Maintenance)
def set_maintenance(
    data: Maintenance,
    user: UserToken = Security(get_user, scopes=["maintenance"]),
) -> Maintenance:
    with SessionFuture() as db:
        maintenance: DbSetting | None = db.execute(
            select(DbSetting).where(DbSetting.name == "Maintenance")
        ).scalar_one_or_none()
        if maintenance is None:
            if data.enabled:
                maintenance = DbSetting(
                    name="Maintenance", setting_type=SettingType.MAINTENANCE_MODE, data={"id": user.id_}
                )
                db.add(maintenance)
        else:
            if not data.enabled:
                db.delete(maintenance)
                maintenance = None
            else:
                if maintenance.data["id"] != user.id_:
                    maintenance.data = {"id": user.id_}
        db.commit()
        return info(maintenance, db)


def info(data: DbSetting | None, db: Session) -> Maintenance:
    if data is None:
        return Maintenance(enabled=False, user="")
    user = db.execute(select(User).where(User.id == data.data["id"])).scalar_one()
    return Maintenance(enabled=True, user=user.name)
