from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

import brewman.schemas.balance_sheet as schemas

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.account_type import AccountType
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...routers.reports.closing_stock import get_closing_stock
from ...routers.reports.profit_loss import get_accumulated_profit
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.BalanceSheet)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["balance-sheet"]),
) -> schemas.BalanceSheet:
    return schemas.BalanceSheet(date_=get_finish_date(request.session), body=[], footer=None)


@router.get("/{date_}", response_model=schemas.BalanceSheet)
def report_data(
    date_: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["balance-sheet"]),
) -> schemas.BalanceSheet:
    with SessionFuture() as db:
        body, footer = build_balance_sheet(datetime.strptime(date_, "%d-%b-%Y"), db)
        set_period(get_start_date(request.session), date_, request.session)
        return schemas.BalanceSheet(date_=date_, body=body, footer=footer)


def build_balance_sheet(date_: date, db: Session) -> tuple[list[schemas.BalanceSheetItem], schemas.BalanceSheetItem]:
    type_list = (
        db.execute(select(AccountType.id).where(AccountType.balance_sheet == True)).scalars().all()  # noqa: E712
    )
    report: list[schemas.BalanceSheetItem] = []
    groups: dict[int, schemas.BalanceSheetItem] = {}
    # Add Net Profit / Loss
    closing_stock = round(get_closing_stock(date_, db), 2)
    net_profit = round(get_accumulated_profit(date_, db), 2) - closing_stock
    total_amount = net_profit
    report.append(
        schemas.BalanceSheetItem(
            name="Net Loss" if net_profit >= 0 else "Net Profit",
            group="",
            amount=Decimal(0),
            sub_amount=round(net_profit, 2),
            order=79000,
        )
    )

    capital_group: AccountType = db.execute(select(AccountType).where(AccountType.name == "Capital")).scalar_one()
    groups[capital_group.id] = schemas.BalanceSheetItem(
        name="",
        group=capital_group.name,
        amount=round(total_amount, 2),
        sub_amount=Decimal(0),
        order=capital_group.order,
    )

    total_amount += closing_stock
    report.append(
        schemas.BalanceSheetItem(
            name="Closing Stock", group="", amount=Decimal(0), sub_amount=round(closing_stock, 2), order=20001
        )
    )

    asset_group: AccountType = db.execute(select(AccountType).where(AccountType.name == "Assets")).scalar_one()
    groups[asset_group.id] = schemas.BalanceSheetItem(
        name="",
        group=asset_group.name,
        amount=round(closing_stock, 2),
        sub_amount=Decimal(0),
        order=asset_group.order,
    )

    amount_sum = func.sum(Journal.amount * Journal.debit)
    query = db.execute(
        select(AccountBase, amount_sum)
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date_ <= date_,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id.in_(type_list),
        )
        .group_by(AccountBase)  # type: ignore
        .order_by(AccountBase.type_id, desc(func.abs(amount_sum)))
    ).all()

    counter = 0
    for account, amount in query:
        # Add Items
        total_amount += amount
        if amount != 0:
            counter += 1
            report.append(
                schemas.BalanceSheetItem(
                    name=account.name,
                    group="",
                    amount=Decimal(0),
                    sub_amount=round(amount, 2),
                    order=account.type_.order + counter,
                )
            )
            if account.type_.id in groups:
                groups[account.type_.id].amount = round(groups[account.type_.id].amount + amount, 2)
            else:
                groups[account.type_.id] = schemas.BalanceSheetItem(
                    name="",
                    group=account.type_.name,
                    amount=round(amount, 2),
                    sub_amount=Decimal(0),
                    order=account.type_.order,
                )

    # Add Subtotals
    for item in groups.values():
        report.append(item)

    footer = schemas.BalanceSheetItem(name="Total", amount=round(total_amount, 2), order=100000)
    return sorted(report, key=lambda d: d.order), footer
