import uuid

from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, HTTPException, Request, Security, status
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session, contains_eager
from sqlalchemy.sql.expression import delete, distinct, func, or_, select, update

import brewman.schemas.closing_stock as schemas

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.batch import Batch
from ...models.closing_stock import ClosingStock
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.product_group import ProductGroup
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.validations import check_journals_are_valid
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken
from .. import get_lock_info


router = APIRouter()


@router.get("", response_model=schemas.ClosingStock)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["closing-stock"]),
) -> schemas.ClosingStock:
    return schemas.ClosingStock(
        date_=get_finish_date(request.session),  # type: ignore
        posted=False,
        cost_centre=schemas.CostCentreLink(id_=CostCentre.cost_centre_purchase()),
        items=[],
    )


@router.get("/{date_}", response_model=schemas.ClosingStock)
def report_data(
    date_: str,
    request: Request,
    d: uuid.UUID | None = None,
    user: UserToken = Security(get_user, scopes=["closing-stock"]),
) -> schemas.ClosingStock:
    set_period(get_start_date(request.session), date_, request.session)
    with SessionFuture() as db:
        return full_report(datetime.strptime(date_, "%d-%b-%Y").date(), d or CostCentre.cost_centre_purchase(), db)


@router.delete("/{date_}")
def delete_voucher(
    date_: str,
    d: uuid.UUID | None = None,
    user: UserToken = Security(get_user, scopes=["closing-stock"]),
) -> None:
    try:
        cost_centre_id = d or CostCentre.cost_centre_purchase()
        with SessionFuture() as db:
            db.execute(
                delete(ClosingStock)
                .where(ClosingStock.date_ == date_, ClosingStock.cost_centre_id == cost_centre_id)
                .execution_options(synchronize_session=False)
            )
            delete_old_vouchers(datetime.strptime(date_, "%d-%b-%Y").date(), cost_centre_id, db)
            db.commit()
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.post("/{date_}", response_model=schemas.ClosingStock)
def post_voucher(
    date_: str,
    d: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["post-vouchers"]),
) -> schemas.ClosingStock:
    try:
        with SessionFuture() as db:
            date_obj = datetime.strptime(date_, "%d-%b-%Y").date()
            vouchers = (
                db.execute(
                    select(Voucher.id).where(
                        Voucher.date_ == date_obj, Voucher.voucher_type == VoucherType.CLOSING_STOCK
                    )
                )
                .scalars()
                .all()
            )
            vouchers = (
                db.execute(
                    select(Journal.voucher_id).where(Journal.cost_centre_id == d, Journal.voucher_id.in_(vouchers))
                )
                .scalars()
                .all()
            )
            db.execute(update(Voucher).where(Voucher.id.in_(vouchers)).values(posted=True, poster_id=user.id_))
            db.commit()
            return full_report(date_obj, d, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def full_report(date_: date, cost_centre_id: uuid.UUID, db: Session) -> schemas.ClosingStock:
    voucher = (
        db.execute(select(Voucher).where(Voucher.date_ == date_, Voucher.voucher_type == VoucherType.CLOSING_STOCK))
        .scalars()
        .first()
    )
    return schemas.ClosingStock(
        date_=date_,
        cost_centre=schemas.CostCentreLink(id_=cost_centre_id),
        items=build_report(date_, cost_centre_id, db),
        creation_date=voucher.creation_date if voucher is not None else None,
        last_edit_date=voucher.last_edit_date if voucher is not None else None,
        user=schemas.UserLink(id_=voucher.user.id, name=voucher.user.name) if voucher is not None else None,
        posted=voucher.posted if voucher is not None else False,
        poster=None
        if voucher is None or voucher.poster is None
        else schemas.UserLink(id_=voucher.poster.id, name=voucher.poster.name),
    )


def build_report(date_: date, cost_centre_id: uuid.UUID, db: Session) -> list[schemas.ClosingStockItem]:
    amount_sum = func.sum(Journal.debit * Inventory.quantity * Inventory.rate * (1 + Inventory.tax)).label("amount")
    quantity_sum = func.sum(Journal.debit * Inventory.quantity).label("quantity")
    query = db.execute(
        select(StockKeepingUnit, quantity_sum, amount_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(
            Voucher.date_ <= date_,
            Journal.cost_centre_id == cost_centre_id,
            or_(Voucher.voucher_type != VoucherType.CLOSING_STOCK, Voucher.date_ != date_),
        )
        .group_by(StockKeepingUnit, Product, ProductGroup)  # type: ignore
        .order_by(ProductGroup.name, Product.name, StockKeepingUnit.units)
    ).all()

    physical_list = (
        db.execute(
            select(ClosingStock).where(ClosingStock.date_ == date_, ClosingStock.cost_centre_id == cost_centre_id)
        )
        .scalars()
        .all()
    )

    ccs = db.execute(
        select(Journal.cost_centre_id, Batch.sku_id)
        .join(Journal.voucher)
        .join(Voucher.inventories)
        .join(Inventory.batch)
        .where(
            Voucher.voucher_type == VoucherType.CLOSING_STOCK,
            Voucher.date_ == date_,
            Journal.cost_centre_id != cost_centre_id,
            Voucher.id.in_(select(Journal.voucher_id).where(Journal.cost_centre_id == cost_centre_id)),
        )
        .group_by(Journal.cost_centre_id, Batch.sku_id)
    ).all()

    body = []
    for sku, quantity, amount in query:
        if quantity != 0 and amount != 0:
            id_ = next((p.id for p in physical_list if p.sku_id == sku.id), None)
            physical = next((p.quantity for p in physical_list if p.sku_id == sku.id), quantity)
            cc = next((schemas.CostCentreLink(id_=c) for (c, s) in ccs if s == sku.id), None)
            body.append(
                schemas.ClosingStockItem(
                    id_=id_,
                    product=schemas.ProductLink(id_=sku.id, name=f"{sku.product.name} ({sku.units})"),
                    group=sku.product.product_group.name,
                    quantity=quantity,
                    amount=amount,
                    physical=physical,
                    cost_centre=cc,
                )
            )
    return body


def get_opening_stock(date_: date, db: Session) -> Decimal:
    opening_stock: Decimal | None = db.execute(
        select(func.sum(Inventory.quantity * Inventory.rate * (1 + Inventory.tax) * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .join(Voucher.inventories)
        .where(Voucher.date_ < date_, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
    ).scalar()
    return Decimal(0) if opening_stock is None else opening_stock


def get_closing_stock(date_: date, db: Session) -> Decimal:
    closing_stock: Decimal | None = db.execute(
        select(func.sum(Inventory.quantity * Inventory.rate * (1 + Inventory.tax) * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .join(Voucher.inventories)
        .where(Voucher.date_ <= date_, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
    ).scalar()
    return Decimal(0) if closing_stock is None else closing_stock


@router.post("", response_model=schemas.ClosingStock)
def save_route(
    request: Request,
    data: schemas.ClosingStock,
    user: UserToken = Security(get_user, scopes=["closing-stock"]),
) -> schemas.ClosingStock:
    try:
        departments = set(i.cost_centre.id_ for i in data.items if i.cost_centre is not None)
        if data.cost_centre.id_ in departments:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Issue department cannot be same as main department",
            )
        if CostCentre.cost_centre_purchase() in departments:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Products cannot be issued to purchase",
            )
        if CostCentre.cost_centre_overall() in departments:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Products cannot be issued to overall",
            )
        if len([i for i in data.items if i.physical > i.quantity]) > 0:
            x = [i for i in data.items if i.physical > i.quantity]
            for y in x:
                print(y)
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Physical Quantity cannot be more than computed quality.",
            )
        if len([i for i in data.items if i.physical != i.quantity and i.cost_centre is None]) > 0:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Products with change in quantity need a department",
            )
        with SessionFuture() as db:
            count_ = db.execute(select(func.count()).where(ClosingStock.date_ == data.date_)).scalar_one()
            if count_ == 0:
                save_cs(data, db)
            else:
                update_cs(data.date_, data.cost_centre.id_, data.items[:], db)
                delete_old_vouchers(data.date_, data.cost_centre.id_, db)
            for dep in departments:
                items = [d for d in data.items if d.cost_centre is not None and d.cost_centre.id_ == dep]
                item = save(data.date_, items, user, db)
                save_inventories(item, items, db)
                amount: Decimal = sum(i.amount for i in item.inventories) or Decimal(0)
                save_journals(item, data.cost_centre.id_, dep, amount, db)
                check_journals_are_valid(item)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return full_report(data.date_, data.cost_centre.id_, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save_cs(data: schemas.ClosingStock, db: Session) -> None:
    for item in data.items:
        cs = ClosingStock(
            date_=data.date_, cost_centre_id=data.cost_centre.id_, sku_id=item.product.id_, quantity=item.physical
        )
        db.add(cs)


def save(date_: date, items: list[schemas.ClosingStockItem], user: UserToken, db: Session) -> Voucher:
    product_accounts: list[uuid.UUID] = list(  # type: ignore
        select(Product.account_id).join(Product.skus).where(StockKeepingUnit.id.in_([i.product.id_ for i in items]))
    )
    account_types: list[int] = list(
        db.execute(select(distinct(AccountBase.type_id)).where(AccountBase.id.in_(product_accounts))).scalars().all()
    )
    allowed, message = get_lock_info([date_], VoucherType.CLOSING_STOCK, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date_=date_,
        narration="",
        is_starred=False,
        user_id=user.id_,
        voucher_type=VoucherType.CLOSING_STOCK,
    )
    db.add(voucher)
    return voucher


def save_inventories(voucher: Voucher, items: list[schemas.ClosingStockItem], db: Session) -> None:
    for item in items:
        batches = list(
            db.execute(
                select(Batch)
                .where(Batch.sku_id == item.product.id_, Batch.quantity_remaining != 0, Batch.name <= voucher.date_)
                .order_by(Batch.name)
            )
            .scalars()
            .all()
        )
        to_issue = item.quantity - item.physical
        while to_issue > 0:
            batch = batches.pop(0)
            quantity_issued = min(to_issue, batch.quantity_remaining)
            inv = Inventory(
                quantity=quantity_issued,
                rate=batch.rate,
                tax=batch.tax,
                discount=batch.discount,
                batch=batch,
            )
            batch.quantity_remaining -= quantity_issued
            to_issue -= quantity_issued
            voucher.inventories.append(inv)
            db.add(inv)


def save_journals(
    voucher: Voucher,
    source: uuid.UUID,
    destination: uuid.UUID,
    amount: Decimal,
    db: Session,
) -> None:
    s = Journal(debit=-1, account_id=AccountBase.all_purchases(), amount=amount, cost_centre_id=source)
    d = Journal(debit=1, account_id=AccountBase.all_purchases(), amount=amount, cost_centre_id=destination)
    voucher.journals.append(s)
    db.add(s)
    voucher.journals.append(d)
    db.add(d)


def update_cs(date_: date, cost_centre_id: uuid.UUID, items: list[schemas.ClosingStockItem], db: Session) -> None:
    old = (
        db.execute(
            select(ClosingStock).where(ClosingStock.date_ == date_, ClosingStock.cost_centre_id == cost_centre_id)
        )
        .scalars()
        .all()
    )
    for i in range(len(old), 0, -1):
        item = old[i - 1]
        index = next((idx for (idx, d) in enumerate(items) if d.id_ == item.id), None)
        if index is not None:
            new_item = items.pop(index)
            if item.quantity != new_item.physical:
                db.execute(update(ClosingStock).where(ClosingStock.id == item.id).values(quantity=new_item.physical))
        else:
            db.execute(delete(ClosingStock).where(ClosingStock.id == item.id))
    for new_item in items:
        cs = ClosingStock(
            date_=date_,
            cost_centre_id=cost_centre_id,
            sku_id=new_item.product.id_,
            quantity=new_item.physical,
        )
        db.add(cs)


def delete_old_vouchers(date_: date, cost_centre_id: uuid.UUID, db: Session) -> None:
    vouchers = (
        db.execute(select(Voucher.id).where(Voucher.date_ == date_, Voucher.voucher_type == VoucherType.CLOSING_STOCK))
        .scalars()
        .all()
    )
    vouchers = (
        db.execute(
            select(func.distinct(Journal.voucher_id)).where(
                Journal.cost_centre_id == cost_centre_id, Journal.voucher_id.in_(vouchers)
            )
        )
        .scalars()
        .all()
    )

    batches = db.execute(select(Inventory.batch_id).where(Inventory.voucher_id.in_(vouchers))).scalars().all()
    db.execute(delete(Journal).where(Journal.voucher_id.in_(vouchers)).execution_options(synchronize_session=False))
    db.execute(delete(Inventory).where(Inventory.voucher_id.in_(vouchers)).execution_options(synchronize_session=False))
    db.execute(delete(Voucher).where(Voucher.id.in_(vouchers)).execution_options(synchronize_session=False))
    for batch in batches:
        db.execute(update(Batch).where(Batch.id == batch).values(quantity_remaining=get_batch_quantity(batch, db)))


def get_batch_quantity(id_: uuid.UUID, db: Session) -> Decimal:
    query = (
        select(func.sum(Inventory.quantity * Journal.debit))
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(Inventory.batch_id == id_, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
    )
    return db.execute(query).scalar_one() or Decimal(0)
