import { AfterViewInit, Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatPaginator } from '@angular/material/paginator';
import { MatSort } from '@angular/material/sort';
import { ActivatedRoute } from '@angular/router';
import { Observable } from 'rxjs';
import { debounceTime, distinctUntilChanged } from 'rxjs/operators';

import { Product } from '../../core/product';
import { ToCsvService } from '../../shared/to-csv.service';

import { ProductListDataSource } from './product-list-datasource';

@Component({
  selector: 'app-product-list',
  templateUrl: './product-list.component.html',
  styleUrls: ['./product-list.component.css'],
})
export class ProductListComponent implements OnInit, AfterViewInit {
  @ViewChild('filterElement', { static: true }) filterElement?: ElementRef;
  @ViewChild(MatPaginator, { static: true }) paginator?: MatPaginator;
  @ViewChild(MatSort, { static: true }) sort?: MatSort;
  list: Product[] = [];
  filter: Observable<string>;
  dataSource: ProductListDataSource;
  form: FormGroup<{
    filter: FormControl<string>;
  }>;

  /** Columns displayed in the table. Columns IDs can be added, removed, or reordered. */
  displayedColumns: string[] = [];

  private _showExtended = true;

  constructor(
    private route: ActivatedRoute,
    private toCsv: ToCsvService,
  ) {
    this.showExtended = true;
    this.form = new FormGroup({
      filter: new FormControl<string>('', { nonNullable: true }),
    });
    this.filter = this.form.controls.filter.valueChanges.pipe(debounceTime(150), distinctUntilChanged());
    this.dataSource = new ProductListDataSource(this.list, this.filter);
  }

  get showExtended(): boolean {
    return this._showExtended;
  }

  set showExtended(value: boolean) {
    this._showExtended = value;
    if (value) {
      this.displayedColumns = ['name', 'costPrice', 'productYield', 'productGroup', 'info'];
    } else {
      this.displayedColumns = ['name', 'costPrice', 'productGroup', 'info'];
    }
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { list: Product[] };

      this.list = data.list;
    });
    this.dataSource = new ProductListDataSource(this.list, this.filter, this.paginator, this.sort);
  }

  ngAfterViewInit() {
    setTimeout(() => {
      if (this.filterElement) {
        this.filterElement.nativeElement.focus();
      }
    }, 0);
  }

  exportCsv() {
    const headers = {
      Code: 'code',
      Name: 'name',
      Units: 'units',
      Fraction: 'fraction',
      FractionUnits: 'fractionUnits',
      CostPrice: 'costPrice',
      Yield: 'yield',
      ProductGroup: 'productGroup',
    };

    const csvData = new Blob([this.toCsv.toCsv(headers, this.dataSource.data)], {
      type: 'text/csv;charset=utf-8;',
    });
    const link = document.createElement('a');
    link.href = window.URL.createObjectURL(csvData);
    link.setAttribute('download', 'products.csv');
    document.body.appendChild(link);
    link.click();
    document.body.removeChild(link);
  }
}
